import itertools
import os
import pydicom
import shutil

from collections import abc
from queue import Empty, Queue
from threading import Thread

from dicomsort import errors, utils
from dicomsort.gui import events


THREAD_COUNT = 2


class Dicom:
    def __init__(self, filename, dcm=None):
        """
        Takes a dicom filename in and returns instance that can be used to sort
        """
        # Be sure to do encoding because Windows sucks
        self.filename = filename

        # Load the DICOM object
        if dcm:
            self.dicom = dcm
        else:
            self.dicom = pydicom.read_file(self.filename)

        self.series_first = False

        self.default_overrides = {
            'ImageType': self._image_type,
            'FileExtension': self._file_extension,
            'SeriesDescription': self._series_description
        }

        self.overrides = dict(self.default_overrides)

    def __getitem__(self, attr):
        """
        Points the reference to the property unless an override is specified
        """
        try:
            item = self.overrides[attr]
            if isinstance(item, abc.Callable):
                return item()

            return item
        except KeyError:
            return getattr(self.dicom, attr)

    def _file_extension(self):
        filename, extension = os.path.splitext(self.dicom.filename)
        return extension

    def _series_description(self):
        if not hasattr(self.dicom, 'SeriesDescription'):
            out = 'Series%04d' % self.dicom.SeriesNumber
        else:
            if self.series_first:
                out = 'Series%04d_%s' % (self.dicom.SeriesNumber,
                                         self.dicom.SeriesDescription)
            else:
                out = '%s_Series%04d' % (self.dicom.SeriesDescription,
                                         self.dicom.SeriesNumber)

        # Strip so we don't have any leading/trailing spaces
        return out.strip()

    def _patient_age(self):
        """
        Computes the age of the patient
        """
        if 'PatientAge' in self.dicom:
            age = self.dicom.PatientAge
        elif 'PatientBirthDate' not in self.dicom or \
                self.dicom.PatientBirthDate == '':
            age = ''
        else:
            age = (int(self.dicom.StudyDate) -
                   int(self.dicom.PatientBirthDate)) / 10000
            age = '%03dY' % age

        return age

    def _image_type(self):
        """
        Determines the human-readable type of the image
        """

        types = {
            'Phase': {'P', },
            '3DRecon': {'CSA 3D EDITOR', },
            'Phoenix': {'CSA REPORT', },
            'Mag': {'FFE', 'M'},
        }

        try:
            image_type = set(self.dicom.ImageType)
        except AttributeError:
            return 'Unknown'

        for typeString, match in types.items():
            if match.issubset(image_type):
                if typeString == '3DRecon':
                    self.dicom.InstanceNumber = self.dicom.SeriesNumber

                return typeString

        return 'Image'

    def get_destination(self, root, directory_format, filename_format):

        # First we need to clean up the elements of directory_format to make
        # sure that we don't have any bad characters (including /) in the
        # folder names
        directory = root
        for item in directory_format:
            try:
                subdir = utils.recursive_replace_tokens(item, self)
                subdir = utils.clean_directory_name(subdir)
            except AttributeError:
                subdir = 'UNKNOWN'

            directory = os.path.join(directory, subdir)

        # Maximum recursion is 5 so we don't end up with any infinite loop
        # situations
        try:
            filename = utils.recursive_replace_tokens(filename_format, self)
            filename = utils.clean_path(filename)
            out = os.path.join(directory, filename)
        except AttributeError:
            # Now just use the initial filename
            origname = os.path.split(self.filename)[1]
            out = os.path.join(directory, origname)

        return out

    def set_anonymization_rules(self, anonymization_lookup):
        # Appends the rules to the overrides so that we can alter them
        if isinstance(anonymization_lookup, dict):
            self.anonymization_lookup = anonymization_lookup
        else:
            raise Exception('Anon rules must be a dictionary')

        if 'PatientBirthDate' in self.anonymization_lookup:
            if self.anonymization_lookup['PatientBirthDate'] != '' or \
                    self.dicom.PatientBirthDate == '':

                self.overrides = dict(
                    self.default_overrides,
                    **anonymization_lookup
                )

                return

            # First we need to figure out how old they are
            if 'PatientAge' in self.dicom and 'StudyDate' in self.dicom:
                self.dicom.PatientAge = self._patient_age()

            if 'StudyDate' in self.dicom:
                # Now set it so it is just the birth year but make it so that
                # the proper age is returned when doing year math
                birth_date = int(self.dicom.PatientBirthDate[4:])
                study_date = int(self.dicom.StudyDate[4:])

                # If the study was performed after their birthday this year
                if study_date >= birth_date:
                    new_birth_date = '%s0101' % self.dicom.PatientBirthDate[:4]
                else:
                    birth_year = self.dicom.PatientBirthDate[:4]
                    new_birth_date = '%d0101' % (int(birth_year) + 1)

                self.anonymization_lookup['PatientBirthDate'] = new_birth_date

        # Update the override dictionary
        self.overrides = dict(self.default_overrides, **anonymization_lookup)

    def is_anonymous(self):
        return self.default_overrides != self.overrides

    def sort(self, root, directory_fields, filename_string, test=False,
             rootdir=None, keep_original=True):

        # If we want to sort in place
        if directory_fields is None:
            destination = os.path.relpath(self.filename, rootdir[0])
            destination = os.path.join(root, destination)
        else:
            destination = self.get_destination(
                root, directory_fields, filename_string
            )

        if test:
            print(destination)
            return

        utils.mkdir(os.path.dirname(destination))

        # Check if destination exists
        while os.path.exists(destination):
            destination = destination + '.copy'

        if self.is_anonymous():
            # Actually write the anonymous data
            # write everything in anonymization_lookup -> Parse it so we can
            # have dynamic fields
            for key in self.anonymization_lookup.keys():
                replacement_value = self.anonymization_lookup[key] % self
                try:
                    self.dicom.data_element(key).value = replacement_value
                except KeyError:
                    continue

            self.dicom.save_as(destination)

            if keep_original is False:
                os.remove(self.filename)

        else:
            if keep_original:
                shutil.copy(self.filename, destination)
            else:
                shutil.move(self.filename, destination)


class Sorter(Thread):
    def __init__(self, queue, output_directory, directory_format,
                 filename_format, lookup=None, keep_filename=False,
                 iterator=None, test=False, listener=None, total=None,
                 root=None, series_first=False, keep_original=True):

        self.directory_format = directory_format
        self.filename_format = filename_format
        self.queue = queue
        self.anonymization_lookup = lookup or dict()
        self.keep_filename = keep_filename
        self.series_first = series_first
        self.keep_original = keep_original
        self.output_directory = output_directory
        self.test = test
        self.iter = iterator
        self.root = root
        self.total = total or self.queue.qsize()

        self.is_gui = False

        if listener:
            self.listener = listener
            self.is_gui = True

        Thread.__init__(self)
        self.start()

    def sort_image(self, filename):
        dcm = utils.isdicom(filename)

        if not dcm:
            return

        dcm = Dicom(filename, dcm)
        dcm.set_anonymization_rules(self.anonymization_lookup)
        dcm.series_first = self.series_first

        # Use the original filename for 3d recons
        if self.keep_filename:
            output_filename = os.path.basename(filename)
        else:
            output_filename = self.filename_format

        dcm.sort(
            self.output_directory,
            self.directory_format,
            output_filename,
            test=self.test,
            rootdir=self.root,
            keep_original=self.keep_original
        )

    def increment_counter(self):
        if self.iter is None:
            return

        count = next(self.iter)

        if self.is_gui is False:
            return

        event = events.CounterEvent(Count=count, total=self.total)
        events.post_event(self.listener, event)

    def run(self):
        while True:
            try:
                filename = self.queue.get_nowait()
                self.sort_image(filename)
                self.increment_counter()
            # TODO: Rescue any other errors and quarantine the files
            except Empty:
                return


class DicomSorter():
    def __init__(self, pathname=None):
        # Use current directory by default
        if not pathname:
            pathname = [os.getcwd(), ]

        if not isinstance(pathname, list):
            pathname = [pathname, ]

        self.pathname = pathname

        self.folders = []
        self.filename = '%(ImageType)s (%(InstanceNumber)04d)%(FileExtension)s'

        self.queue = Queue()

        self.sorters = list()

        # Don't anonymize by default
        self.anonymization_lookup = dict()

        self.keep_filename = False
        self.series_first = False
        self.keep_original = True

    def is_sorting(self):
        for sorter in self.sorters:
            if sorter.is_alive():
                return True

        return False

    def set_anonymization_rules(self, anonymization_lookup):
        # Appends the rules to the overrides so that we can alter them
        if not isinstance(anonymization_lookup, dict):
            raise Exception('Anon rules must be a dictionary')

        self.anonymization_lookup = anonymization_lookup

    def folder_format(self):
        # Check to see if we are using the origin directory structure
        if self.folders is None:
            return None

        # Make a local copy
        folder_list = self.folders[:]

        return folder_list

    def sort(self, output_directory, test=False, listener=None):
        # This should be moved to a worker thread
        for path in self.pathname:
            for root, _, files in os.walk(path):
                for filename in files:
                    self.queue.put(os.path.join(root, filename))

        number_of_files = self.queue.qsize()
        dir_format = self.folder_format()

        self.sorters = list()

        iterator = itertools.count(1)

        for _ in range(min(THREAD_COUNT, number_of_files)):
            sorter = Sorter(
                self.queue, output_directory, dir_format, self.filename,
                self.anonymization_lookup, self.keep_filename,
                iterator=iterator, test=test, listener=listener,
                total=number_of_files, root=self.pathname,
                series_first=self.series_first,
                keep_original=self.keep_original
            )

            self.sorters.append(sorter)

    def available_fields(self):
        for path in self.pathname:
            for root, dirs, files in os.walk(path):
                for filename in files:
                    dcm = utils.isdicom(os.path.join(root, filename))
                    if dcm:
                        return dcm.dir('')

        msg = ''.join([';'.join(self.pathname), ' contains no DICOMs'])
        raise errors.DicomFolderError(msg)



import os

from os.path import expanduser

configuration_version = '2.0'

default_filename = '%(ImageType)s (%(InstanceNumber)04d)%(FileExtension)s'

default_configuration = {
    'Version': configuration_version,
    'Anonymization': {
        'Fields': [
            'OtherPatientsIDS',
            'PatientID',
            'PatientBirthDate',
            'PatientName',
            'ReferringPhysiciansName',
            'RequestingPhysician'
        ],
        'Replacements': {
            'PatientName': 'ANONYMOUS',
            'PatientID': '%(PatientName)s'
        }
    },
    'FilenameFormat': {
        'FilenameString': default_filename,
        'Selection': 0
    },
    'Miscpanel': {
        'KeepSeries': 'True',
        'SeriesFirst': 'False',
        'KeepOriginal': 'True'
    }
}

filename = 'dicomSort.ini' if os.name == 'nt' else '.dicomSort.ini'
configuration_file = os.path.join(expanduser('~'), filename)



import wx

from wx import html
from wx.lib.agw import multidirdialog


class HtmlWindow(html.HtmlWindow):
    def __init__(self, parent, id, size):
        wx.html.HtmlWindow.__init__(self, parent, id, size=size)
        if "gtk2" in wx.PlatformInfo:
            self.SetStandardFonts()


class MultiDirDlg(multidirdialog.MultiDirDialog):
    """
    We have to over-ride this because of the IndexError on win7
    """

    def __init__(self, *args, **kwargs):
        super(MultiDirDlg, self).__init__(*args, **kwargs)

    def SetupDirCtrl(self, *args, **kwargs):
        try:
            super(MultiDirDlg, self).SetupDirCtrl(*args, **kwargs)
        except IndexError:
            self.folderText.SetValue('')



import dicomsort
import os
import traceback
import webbrowser
import wx

from urllib.parse import quote

from dicomsort.gui import icons
from dicomsort.gui.help import helpHTML
from dicomsort.gui.overrides import HtmlWindow
from wx.adv import AboutBox, AboutDialogInfo
from wx.lib.agw import hyperlink


meta = dicomsort.Metadata


class AboutDlg:

    def __init__(self, parent=None):
        self.info = AboutDialogInfo()
        self.info.SetIcon(icons.about.GetIcon())

        self.info.SetName(meta.pretty_name)
        self.info.SetVersion(meta.version)

        self.info.SetCopyright(meta.copyright)
        self.info.SetWebSite(meta.website)

        self.GenerateDescription()

        AboutBox(self.info, parent=parent)

    def GenerateDescription(self):
        description = ("      Program designed to sort DICOM images       \n" +
                       "     into directories based upon DICOM header     \n" +
                       "      information. The program also provides      \n" +
                       "       additional functionality such as the       \n" +
                       "anonymization of DICOM images for patient privacy.")

        self.info.SetDescription(description)


class CrashReporter(wx.Dialog):
    title = 'DICOM Sort Error'

    def __init__(self, parent, **kwargs):
        super(CrashReporter, self).__init__(
            parent, -1, self.title, size=(400, 400)
        )

        self.type = kwargs.pop('type', None)
        self.value = kwargs.pop('value', None)
        self._traceback = kwargs.pop('traceback', None)

        self.create()
        self.SetIcon(icons.main.GetIcon())

    def traceback(self):
        lines = traceback.format_exception(type, self.value, self._traceback)
        return '\n'.join(lines)

    def body(self):
        return \
            '#### Describe the Issue:\n' + \
            'A clear and concise description of what the bug is.\n\n' + \
            '#### Expected Behavior:\n' + \
            'What you expected to happen.\n\n' + \
            '#### Steps to Reproduce:\n' + \
            'How to reproduce the issue.\n\n' + \
            '## Environment\n' + \
            '#### DICOM Sort Version:\n{}\n\n'.format(meta.version) + \
            '#### Operating System:\n{}\n\n'.format(' '.join(os.uname())) + \
            '#### Traceback:\n' + \
            '```\n' + \
            self.traceback() + \
            '\n```\n'

    def create(self):
        vbox = wx.BoxSizer(wx.VERTICAL)

        text = ''.join([
            'Dicom Sort had a problem and crashed. ',
            'In order to help diagnose the problem, you can submit an ',
            'issue and include the traceback below:',
        ])

        static = wx.StaticText(self, -1, text)
        static.Wrap(380)
        vbox.Add(static, 0, wx.ALL, 10)

        value = '\n'.join(
            traceback.format_exception(type, self.value, self._traceback)
        )
        self.comments = wx.TextCtrl(self, style=wx.TE_MULTILINE, value=value)
        vbox.Add(self.comments, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 10)

        hbox = wx.BoxSizer(wx.HORIZONTAL)

        self.submit = wx.Button(self, -1, 'File Issue')
        self.submit.Bind(wx.EVT_BUTTON, self.on_file)

        self.done = wx.Button(self, -1, 'Done')
        self.done.Bind(wx.EVT_BUTTON, self.on_button)

        hbox.Add(self.submit, 0, wx.RIGHT, 10)
        hbox.Add(self.done, 0, wx.RIGHT, 10)

        vbox.Add(hbox, 0, wx.TOP | wx.BOTTOM | wx.ALIGN_RIGHT, 10)

        self.SetSizer(vbox)

    def on_file(self, *_event):
        webbrowser.open(meta.issue_url + '?body=' + quote(self.body()))

    def on_button(self, event):
        if self.IsModal():
            self.EndModal(event.EventObject.Id)
        else:
            self.Close()


class HelpDlg(wx.Dialog):

    def __init__(self, parent=None, **kwargs):

        super(HelpDlg, self).__init__(
            parent, -1, '{} Help'.format(meta.pretty_name),
            style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER | wx.TAB_TRAVERSAL
        )

        self.hwin = HtmlWindow(self, -1, size=(400, 200))

        self.hwin.SetPage(helpHTML)
        irep = self.hwin.GetInternalRepresentation()

        self.hwin.SetSize((irep.GetWidth(), int(irep.GetHeight() / 4)))
        self.Show()

        self.SetClientSize(self.hwin.GetSize())
        self.CenterOnParent(wx.BOTH)
        self.SetFocus()

        self.Bind(wx.EVT_CLOSE, self.close)

    def close(self, *_):
        self.Destroy()


class SeriesRemoveWarningDlg(wx.Dialog):

    def __init__(self, parent, id=-1, size=(300, 200), config=None):
        wx.Dialog.__init__(
            self, parent, id, 'Remove Series Description?', size=size
        )

        vbox = wx.BoxSizer(wx.VERTICAL)

        inputText = ''.join(['Are you sure you want to remove the default\n',
                             'attribute SeriesDescription?\n\n',
                             'This may cause filename conflicts unless you\n',
                             'use the original filenames.'])

        txt = wx.StaticText(self, -1, inputText, style=wx.ALIGN_CENTER)

        change = wx.Button(
            self, -1, 'Yes. Use original Filenames', size=(-1, 20)
        )

        accept = wx.Button(
            self, -1, 'Yes. Use Custom Filenames', size=(-1, 20)
        )
        cancel = wx.Button(self, -1, 'Cancel', size=(-1, 20))

        change.Bind(wx.EVT_BUTTON, self.OnChange)
        accept.Bind(wx.EVT_BUTTON, self.OnAccept)
        cancel.Bind(wx.EVT_BUTTON, self.OnCancel)

        self.Bind(wx.EVT_CLOSE, self.OnCancel)

        vbox.Add(txt, 1, wx.ALL | wx.ALIGN_CENTER, 10)
        vbox.Add(change, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        vbox.Add(accept, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        vbox.Add(cancel, 0, wx.ALL | wx.ALIGN_CENTER, 5)

        self.SetSizer(vbox)

        self.Destroy()

    def OnChange(self, *evnt):
        self.choice = 1
        self.Destroy()

    def OnCancel(self, *evnt):
        self.choice = 0
        self.Destroy()

    def OnAccept(self, *evnt):
        self.choice = 2
        self.Destroy()


class QuickRenameDlg(wx.Dialog):

    def __init__(self, *args, **kwargs):
        tmp = kwargs.pop('anonList')
        wx.Dialog.__init__(self, *args, **kwargs)
        self.anonList = tmp
        self.values = dict()

        vbox = wx.BoxSizer(wx.VERTICAL)

        self.values = self.anonList.GetReplacementDict()
        if 'PatientName' in self.values:
            initial = self.values['PatientName']
        else:
            initial = ''

        txt = wx.StaticText(self, -1, 'PatientName')
        vbox.Add(txt, 0, wx.TOP | wx.ALIGN_CENTER, 15)

        self.patientName = wx.TextCtrl(self, -1, initial, size=(200, 20),
                                       style=wx.TE_PROCESS_ENTER)

        self.Bind(wx.EVT_TEXT_ENTER, self.OnAccept, self.patientName)

        vbox.Add(self.patientName, 0, wx.TOP | wx.ALIGN_CENTER, 5)

        self.samecheck = wx.CheckBox(self, -1, 'Use as Patient ID')
        self.samecheck.SetValue(True)

        vbox.Add(self.samecheck, 0, wx.TOP | wx.ALIGN_CENTER, 10)

        self.btnOK = wx.Button(self, -1, 'Ok')
        self.btnOK.Bind(wx.EVT_BUTTON, self.OnAccept)

        vbox.Add(self.btnOK, 0, wx.TOP | wx.ALIGN_CENTER, 15)

        self.SetSizer(vbox)

        self.patientName.SetFocus()

    def GetValues(self):
        res = dict()
        res['PatientName'] = self.patientName.GetValue()

        if self.samecheck.IsChecked():
            res['PatientID'] = '%(PatientName)s'

        return res

    def OnAccept(self, *evnt):
        oldDict = self.anonList.GetReplacementDict()
        oldDict.update(self.GetValues())
        self.anonList.SetReplacementDict(oldDict)
        self.Close()


class UpdateDlg(wx.Dialog):
    def __init__(self, parent, version):
        super(UpdateDlg, self).__init__(parent, size=(300, 170), style=wx.OK)
        message = ''.join([
            'A new version of {} is available.\n'.format(meta.pretty_name),
            'You are running Version %s and the newest is %s.\n'
        ])

        vbox = wx.BoxSizer(wx.VERTICAL)

        head = wx.StaticText(
            self, -1, label='Update Available', style=wx.ALIGN_CENTER
        )
        head.SetFont(wx.Font(12, wx.DEFAULT, wx.NORMAL, wx.BOLD))

        vbox.Add(head, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP, 15)

        txt = wx.StaticText(
            self, -1, label=message % (dicomsort.__version__, version),
            style=wx.ALIGN_CENTER)
        vbox.Add(txt, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.ALL, 5)

        self.link = hyperlink.HyperLinkCtrl(self, -1)
        self.link.SetURL(URL=meta.website)
        self.link.SetLabel(label='Click here to obtain the update')
        self.link.SetToolTip(meta.website)
        self.link.AutoBrowse(False)

        self.Bind(hyperlink.EVT_HYPERLINK_LEFT, self.OnUpdate, self.link)

        vbox.Add(self.link, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.ALL, 0)

        ok = wx.Button(self, -1, "OK")
        vbox.Add(ok, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.ALL, 15)
        self.Bind(wx.EVT_BUTTON, self.OnClose, ok)

        self.SetSizer(vbox)
        self.CenterOnParent()

    def OnUpdate(self, *evnt):
        self.link.GotoURL(self.link.GetURL())
        self.Destroy()

    def OnClose(self, *evnt):
        self.Destroy()



from dicomsort.gui import widgets


class AnonymizeList(widgets.CheckListCtrl):

    def __init__(self, *args, **kwargs):
        super(AnonymizeList, self).__init__(*args, **kwargs)

        self.InsertColumn(0, 'DICOM Property', width=200)
        self.InsertColumn(1, 'Replacement Value', width=-1)

        self.SetColumnEditable(1)

    def GetReplacementDict(self):
        res = dict()

        x = [i for i in range(self.ItemCount) if len(self.GetStringItem(i, 1))]

        for row in x:
            res[self.GetStringItem(row, 0)] = self.GetStringItem(row, 1)

        return res

    def GetAnonDict(self):
        return {k: v for k, v in self.GetCheckedStrings()}

    def SetReplacementDict(self, dictionary):
        keys = list(dictionary.keys())
        indices = self.FindStrings(keys, 0)

        for i, row in enumerate(indices):
            if row is None:
                continue

            self.SetItem(row, 1, dictionary[keys[i]])

    def CheckStrings(self, strings, col=0):
        for index in self.FindStrings(strings, col):
            if index is None:
                continue

            self.CheckItem(index)

    def GetDicomField(self, row):
        return self.GetItem(row, 0).Text



import dicomsort
import re
import wx

from threading import Thread
from urllib.request import urlopen

from dicomsort.gui import events

meta = dicomsort.Metadata


def version_tuple(version):
    parts = list()

    for component in version.split('.'):
        number_match = re.match('^\\d+', component)

        if number_match is None:
            continue

        parts.append(int(number_match.group()))

    return tuple(parts)


VERSION_TUPLE = version_tuple(dicomsort.__version__)


def latest_version():
    try:
        f = urlopen(meta.version_url)
        version = f.read().rstrip()

        if '404' in version:
            return None

        return version
    except IOError:
        return None


def update_available():
    # First try to see if we can connect
    latest = latest_version()

    if latest is None or latest == dicomsort.__version__:
        return None

    # Break it up to see if it's a dev version
    latest_version_tuple = version_tuple(latest)
    for ind in range(len(latest_version_tuple)):
        if latest_version_tuple[ind] > VERSION_TUPLE[ind]:
            return latest
        return None


class UpdateChecker(Thread):

    def __init__(self, frame, listener):
        self.frame = frame
        self.listener = listener

        Thread.__init__(self)
        self.name = 'UpdateThread'

        # Make it a daemon so that when the MainThread exits, it is terminated
        self.daemon = True
        self.start()

    def run(self):
        version = update_available()

        if version:
            print('Current Version is %s' % version)
            # Send an event to the main thread to tell them
            event = events.UpdateEvent(version=version)
            wx.PostEvent(self.listener, event)



import dicomsort
import os

from setuptools import find_packages, setup

metadata = dicomsort.Metadata

if __name__ == '__main__':
    setup(
        **metadata.to_dict(),
        packages=find_packages(),
        install_requires=[
            'configobj >= 5.0.0',
            'pydicom >= 2.0.0',
            'wxpython >= 4.0.0',
        ],
        zip_safe=False,
        scripts=[
            os.path.join('bin', 'dicomsort')
        ],
    )
